#!/usr/bin/perl -w
#   Test program for first test of servo movement
#   Copyright (C) <2017>  <Rolf Jethon>
#   Version 1.1
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

use lib '/usr/local/bin/Modules';
use vars qw/$api $j $resfactor $dev @serial_content $stepwidth $servores $set1 $set2/;
use Device::SerialPort qw( :PARAM :STAT 0.07 );
use Device::PWMGenerator::PCA9685;
use ConfigL;
use WiringPi::API qw(:wiringPi);
$api=WiringPi::API->new;
$api->setup; # use wiringpi port numbers
use warnings;
use Time::HR;
my $waittime_serial=8000000;                          # wait for 8 ms (8000000 ns)
my $currentservo=0;
my $secondservo=1;
init_ports();
print "\n";
sleep 10;
run_loop();
#############################################################
#  checks the input states and runs the mp3s according to the 
#  key sets
#############################################################
sub run_loop {
  while (1) {                                         # keep the loop always running
    get_one_read();
    my $next=$serial_content[3];
    my $prev=$serial_content[2];
    if ($next==0) {
      if ($currentservo < $ConfigL::num_servos - 1) {
        $currentservo++;
        while ($next==0) {                            # wait until the button is released
          get_one_read();                             # read one serial set
          $next=$serial_content[3];
          put_one_move();
          hsleep (20000000);                          # do nothing for 20 ms
        }
        next;
      }
    }
    if ($prev==0) {
      if ($currentservo > 0) {
        $currentservo--;
        while ($prev==0) {                            # wait until the button is released
          get_one_read();                             # read one serial set
          $prev=$serial_content[2];
          put_one_move();
          hsleep (20000000);                          # do nothing for 20 ms
        }
        next;
      }
    }
    put_one_move();
    hsleep (20000000);                                # do nothing for 20 ms
  }
}
#############################################################
#  move one set - send positions to servos
#############################################################
sub put_one_move {
  $secondservo=$currentservo+1;
  my $pos1=int($serial_content[4]/($ConfigL::servores/($$ConfigL::servosettings[$currentservo][4] - $$ConfigL::servosettings[$currentservo][3]))+$$ConfigL::servosettings[$currentservo][3]);
  $dev->setChannelPWM($currentservo,0,$pos1);
  chop $serial_content[3];
  chop $serial_content[3];
  print "\fX-pos(0-$ConfigL::servores)	|Joystick X 	|Servo $currentservo\n";
  print "-----------------------------------------------\n";
  print "$serial_content[4]		|$serial_content[0]		| $pos1\n";
  &$pwm_en(0);
}
#############################################################
#  initialize all ports
#############################################################
sub init_ports {
  $dev = Device::PWMGenerator::PCA9685->new(
    I2CBusDevicePath => $ConfigL::i2cport,           # the i2c device we send commands to 
    I2CDeviceAddress => $ConfigL::i2c_address,
    debug            => 0,
    frequency        => $ConfigL::i2c_freq, #Hz
  );
  $dev->enable();
  # init pwm enable pin 11 ($pwm_en)as output and set to 1 (inactive)
  $api->pin_mode(0,1);                               # port 0 (pin 11) as output
  $api->write_pin(0,1);                              # init port 0 to 1
  $pwm_en=sub{$api->write_pin(0,shift)};
  #$pwm_en=sub{Gpio::set('PA1',shift)};
  $serial = Device::SerialPort->new($ConfigL::serialport); # init the serial port at pin 10 (RXD)
  $serial->baudrate(38400);
  $serial->databits(8);
  $serial->stopbits(1);
  $serial->purge_all();
  $serial->rts_active(0);
  $serial->dtr_active(0);
  $serial->purge_all();                              # flush the buffer

}
#############################################################
#  a hires sleep ( in nanoseconds )
#############################################################
sub hsleep {
  my $duration=shift;
  $now=gethrtime();
  while ($now + $duration >= gethrtime()) {           # As long as the sleep duration is not reached, keep in loop
  }
  return 1;
}
#############################################################
#  reads the serial line as long as we get a string, followed by \n
#  returns the content splittet into an array
#############################################################
sub get_one_read {
  my ($count,$data,$i)=(0,0,0);
  while (! (substr $data,-1 eq "\n")) {
    $serial->write('4');              # sent the command to the arduino to send one set
    hsleep ($waittime_serial);
    ($count,$data)=$serial->read(32);
    if ( $count ) {                   # if we have data, put it into the array
      @serial_content=split / /,$data;
      $serial_content[4]=int($ConfigL::servores * ($serial_content[0]-$ConfigL::joystick_x_start) / ($ConfigL::joystick_x_end-$ConfigL::joystick_x_start));
      if ( $serial_content[4] < 0 ) { $serial_content[4]=0; }
    } else { $serial->purge_all(); }  # else try again
    if ($i >= 10) { die "serial device does not respond\n";}
    $i++;
  }
}
