#!/usr/bin/perl -w
#------------------------------------------------------------
#   Program to aqire living thing movements according to the MP3 to be played
#   Copyright (C) <2017>  <Rolf Jethon>
#   Version 1.1
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#------------------------------------------------------------
use strict;
use lib '/usr/local/bin/Modules';
use ConfigL;
use Curses::UI;
use Device::SerialPort qw( :PARAM :STAT 0.07 );
use WiringPi::API qw(:wiringPi);
use Time::HR;
use Device::PWMGenerator::PCA9685;
use Audio::Play::MPG123;
use File::Find::Rule;
use vars qw/$frames $full_range $maxstep $minstep $nummoves $steps_determined $saved @active_x @channely @channelx @active_y $justview $api $serial $filepos $mp3 $maxfiles $recording $just_loaded $outfilename $inext $iprev $nextp $prevp $outfilename $servo @servocontent $dual $previouscount $contentcount $pwm_en $dev $periodstart @filelist @serial_content/;
my $mp3dir=$ARGV[0];                                  # the name of the MP3 dir to process
if ( ! $mp3dir ) {                                    # stop if no argument has been passed
  print "usage: $0 <mp3_dirname>\n";
  exit 0;
}
system 'clear';
$api = WiringPi::API->new;
$api->setup; # use wiringpi port numbers
my $player = new Audio::Play::MPG123;
init_ports();
my $cui = new Curses::UI( -color_support => 1);
my $sectionheight;
if ($ConfigL::num_servos < 6) { $sectionheight = 8 } else { $sectionheight = $ConfigL::num_servos+2 } 
my $statuspos=0;                     # the actual position in the status section to be output to
my @statuslines;                     # keeps the status messages (ring buffer)
my $filepath='';
# ----------- create the section frames -----------
my $win = $cui->add(                 # The base window 
                    'win1', 'Window',
                     -border => 0,
                     -y    => 0,
                     -bfg  => 'red',
                     -title => 'Tracking moves and save to servo file',
                   );
my $win0 = $win->add(                # The file frame
                    'fra0', 'Container',
                     -border => 1,
                     -y    => 0,
                     -bfg  => 'red',
                     -title => 'File:',
                     -height => 4,
                     -releasefocus =>1
                   );
my $win1 = $win->add(               # The servo selection frame for joystick X
                    'fra1', 'Container',
                     -border => 1,
                     -y    => 4,
                     -bfg  => 'red',
                     -title => 'Channel X',
                     -height => $sectionheight,
                     -width => 31,
                     -releasefocus =>1
                   );
my $win1a = $win->add(              # The servo selection frame for joystick Y
                    'fra1a', 'Container',
                     -border => 1,
                     -y    => 4,
                     -x    => 31,
                     -bfg  => 'red',
                     -title => 'Channel Y',
                     -height => $sectionheight,
                     -width => 31,
                     -releasefocus =>1
                   );
my $win2 = $win->add(               # The tracking range selection frame
                     'fra2', 'Container',
                     -border => 1,
                     -y    => 4,
                     -x    => 62,
                     -bfg  => 'red',
                     -title => 'Track Sets:',
                     -height => $sectionheight,
                     -releasefocus =>1
                   );
my $win3 = $win->add(               # The status frame
                    'fra3', 'Container',
                     -border => 1,
                     -y    => $sectionheight+4,
                     -bfg  => 'red',
                     -title => 'Status:',
                     -height => 7,
                     -focusable =>0
                   );
my $win4 = $win->add(               # The action frame
                    'fra4', 'Container',
                     -border => 1,
                     -y    => $sectionheight+11,
                     -bfg  => 'red',
                     -title => 'Process:',
                     -height => 3,
                     -releasefocus =>1
                   );

$cui->set_binding(\&exit_dialog,"\cQ"); 
# ----------- setup the file section ---------
my $filelabel = $win0->add('myfilelabel', 'Label', -text=>"Path: ", -bold=>1);
my $fileentry = $win0->add('mytextentry', 'TextEntry', -text=>$filepath, -sbborder=>1,-y=>0,-x=>6);
my $nextbutton = $win0->add('mynextbutton','Buttonbox', -buttons => [
                            { -label => '<<', -value=>4, -onpress=> \&mpprev10},
                            { -label => 'Prev', -value=>2, -shortcut=>'p',-onpress=> \&mpprev},
                            { -label => 'Next', -value=>1, -shortcut=>'n',-onpress=> \&mpnext},
                            { -label => '>>', -value=>5, -onpress=> \&mpnext10},
                            { -label => 'Save', -value=>3, -shortcut=>'s',-onpress=> \&savesv}
                          ],-y=>1,-fg=>'red',-bg=>'black');
# ----------- Set up the servo selection ------
for ( my $servonr=0;$servonr < $ConfigL::num_servos;$servonr++) {
  $channelx[$servonr] = $win1->add("mychannelx$servonr",'Checkbox', 
                                   -label=>"$$ConfigL::servosettings[$servonr][5]($servonr)", 
                                   -y=>$servonr,
                                   -onchange=>\&unchecky
                                  );
}
for ( my $servonr=0;$servonr < $ConfigL::num_servos;$servonr++) {
  $channely[$servonr] = $win1a->add("mychannely$servonr",'Checkbox',
                                    -label=>"$$ConfigL::servosettings[$servonr][5]($servonr)",
                                    -y=>$servonr,
                                    -onchange=>\&uncheckx
                                  );
}
# ------------- define the movement record range ------
my $startsetlabel = $win2->add('mystartsetlabel', 'Label', -text=>"Start at set: ", -bold=>1);
my $startset = $win2->add('mystartset', 'TextEntry', -sbborder=>1,-y=>1);
my $stopsetlabel = $win2->add('mystopsetlabel', 'Label', -text=>"Stop at set: ", -bold=>1,-y=>2);
my $stopset = $win2->add('mystopset', 'TextEntry', -sbborder=>1,-y=>3);
my $maximize = $win2->add('mymaximize', 'Checkbox', -label=>'Update',-checked=>1,-y=>4);
my $fullrange = $win2->add('myfullrange', 'Checkbox', -label=>'Full range',-checked=>1,-y=>5);
# ------------ setup the status section ---------------
my $status = $win3->add('statustext','TextViewer',-fg=>'yellow');
my $statusheight=$status->height();
my $proclabel = $win4->add('myproclabel', 'Label', -text=>" Exit program with Ctrl-Q ", -x=>30, -bg=>'black',-fg=>'white');
# ----------- setup the actions to perform ------------
my $procbutton = $win4->add('myprocbutton','Buttonbox', -buttons => [
                            { -label => 'Record', -value=>1, -shortcut=>'r',-onpress=>\&record},
                            { -label => 'View', -value=>2, -shortcut=>'v',-onpress=>\&view}
                          ],-fg=>'red',-bg=>'black');
$fileentry->focus();
########################################################
@_= File::Find::Rule->file()                          # check the mp3 directory and fill the array with the file names
                           ->name("*.mp3")
                           ->maxdepth(1)
                           ->in( $mp3dir );
@filelist=sort(@_);
$maxfiles=$#filelist;                                 # note the number of mp3s in folder
if ( $maxfiles == -1) { err("No files to process in $mp3dir")}
#hardwaretest();
my $stepwidth=50000000;                               # use a step width of 50 ms (50000000 ns)
my $waittime_serial=8000000;                          # wait for 8 ms (8000000 ns)
$mp3=get_mp(-1);                                      # determine the first mp3 file to read
printstatus("$maxfiles MP3 files in folder $mp3dir\n");
load_file();
maxrange();                                           # after loading use the maximum moves
$cui->mainloop();
########################################################
# exit the script
########################################################
sub exit_dialog{
  $cui->mainloopExit();
  exec 'reset';
}

########################################################
# start recording
########################################################
sub record {
  $cui->leave_curses();
  $justview=0;
  for (my $a=0;$a<$ConfigL::num_servos;$a++){
    $active_x[$a]=$channelx[$a]->get();              # make a shortcut for quick access to the checkbox content
    $active_y[$a]=$channely[$a]->get();
  } 
  $full_range=$fullrange->get();
  $minstep=$startset->get();                         # prepare startstep
  $maxstep=$stopset->get();                         # and stopstep
  system 'clear';
  print "Press start button on joystick to start recording                               \n";
  print "--------------------------------------------------------------------------------\n";
  wait_for_start();
  if ($maximize->get()==1) {                        # if we want update of the max step field
    if (! $steps_determined==1){                    # if no new capture after read of mp3
      $stopset->text($contentcount);
    }
    else {
      $stopset->text($previouscount);               # with already captured set to previous maximum steps - partial read does not set new maximum
    }
  }
  $cui->reset_curses();
}
sub view {
  $cui->leave_curses();
  $justview=1;
  undef @active_x;
  undef @active_y;
  $full_range=$fullrange->get();
  $minstep=$startset->get();                         # prepare startstep
  $maxstep=$stopset->get();                         # and stopstep
  system 'clear';
  print "Press start button on joystick to start viewing                                 \n";
  print "--------------------------------------------------------------------------------\n";
  wait_for_start();
  $saved=1;
  $cui->reset_curses();
}
#############################################################
#  waits for a button to be pressed on the steering stick
#  returns the content splittet into an array
#############################################################
sub wait_for_start {
  my $stop=1;
  my $start;
mainloop: while ($stop==1) {
    get_one_read();                                   # read one serial set
    $stop=$serial_content[3];        
    $start=$serial_content[2];        
    if ($start == 0) {                                # check for the start button
      $saved=0;                                       # indicate that changes have not been saved
      while ($start==0) {                             # keep in loop until start button is released
        get_one_read();                               # read one serial set
        $start=$serial_content[2];        
      }
      $player->load($mp3);
#---------------------------------------------------------------------------------------------------------------------------------------
      $frames=$player->frame();                       # Need this nasty construction to convince Audio::Play::MPG123 providing the frame 
      $player->poll(1);                               # count, since it does not deliver the frame count when the file has not start
      $frames=$player->frame();                       # playing ...
      $player->load($mp3);                            # make sure to start at the beginning of the file
#---------------------------------------------------------------------------------------------------------------------------------------
      if ($full_range==1){                             # record during whole MP3 or start with offset
        $contentcount=0;
      } else {
        my $jump=int(($frames->[0]+$frames->[1])*$minstep/($previouscount));
        $player->jump ($jump);
        $contentcount=$minstep;
      }
      $periodstart=gethrtime();                       # synchronize time reference with the start of the MP3
      $player->poll(0);
      $recording=1;                                   # indicator, that recording is active
      add_one_set();                                  # add the first set from the previous reading (start key trigger)
    }
    if ( $recording==1 && (($periodstart + $stepwidth) <= gethrtime())) {
      get_one_read();                                 # if the period time has reached, read the cross stick
      #add_one_set();                                 # and add the data to the array -> not necessary, because after read
    }                                                 # recording is set to 2 by hsleep
    if ( $recording==2 ) {
      add_one_set();                                  # if the period is reached during a wait, take the data from the recent read
      if ((! $full_range==1) && $contentcount > $maxstep ){
        $recording=0;
        print "Press stop button to return to menu or start to run again\n";
        next mainloop;
      } 
      $recording=1;                                   # reset the "period end during wait" flag 
    }
  }
  $recording=0;
}
#############################################################
#  add one data set to the array
#############################################################
sub add_one_set {
  print "$contentcount "; 
  for (my $svo=0;$svo<$ConfigL::num_servos;$svo++){
    if ($justview==1){ print"\[$svo:$servocontent[$contentcount][$svo]\]"}
    else {
      if ($active_x[$svo]) { $servocontent[$contentcount][$svo]=$serial_content[4];print "\[$svo:$serial_content[4]\]"}
      if ($active_y[$svo]) { $servocontent[$contentcount][$svo]=$serial_content[5];print "\[$svo:$serial_content[5]\]"}
    }
  }
  print "\n";
  for (my $c=0;$c<$ConfigL::num_servos;$c++) {
    if ( $contentcount==0 && $servocontent[$contentcount][$c]==0) { 
      $servocontent[$contentcount][$c]=$ConfigL::servores/2;   # if no data available, use middle position
    } else {
      if ( $servocontent[$contentcount][$c]==0) { $servocontent[$contentcount][$c]=$servocontent[$contentcount-1][$c]} 
    }                                                 # fill empty fields with the content of the previous set
  } 
  put_one_move($servocontent[$contentcount]);
  $periodstart+=$stepwidth;
  $contentcount++; 
}
#############################################################
#  a hires sleep ( in nanoseconds )
#############################################################
sub hsleep {
  my $duration=shift;
  my $now=gethrtime();
  while ($now + $duration >= gethrtime()) {           # As long as the sleep duration is not reached, keep in loop
    if ($recording > 0) {
      $player->poll(0);
      if ($player->state == 0) {
         $recording=0;                                # indicate to quit from wait_for_start
         if ($steps_determined!=1 && $justview!=1 ) {                # determine the max lenght only once
           $previouscount=$contentcount;              # note the maximum duration of the mp3 
           $steps_determined=1;
         } #else {
           #if ($contentcount<$previouscount){$previouscount=$contentcount} # make sure the shortest maxima applies ,to prevent
         #}                                                                 # rapid unwanted movement at the end of the MP3
         print "Press stop button to return to menu or start to run again\n" # if the MP3 is finished -> loop
      } else {                                              
        if ($periodstart + $stepwidth <= gethrtime()) {$recording=2} # tell that the period time has reached
      }
    }
  }
  return;
}
#############################################################
#  load previously recorded data from file
#############################################################
sub load_file{
  $outfilename=$mp3;
  $outfilename=~s/\.mp3//i;
  $outfilename=$outfilename.'.svo';
  showpath($outfilename);                             # show the path in file section
  printstatus( "Servo output filename is $outfilename\n");
  my $exist = (stat("$outfilename"))[2];              # check if a servo file exists
  $steps_determined=0;                                # note that never a mp3 end has been reched
  if ( $exist ) {                                     # load it, if so
    my $data;
    open (BIN,"<$outfilename")||err("Cannot open File $outfilename $!");
    while (read BIN,my $chunk,2048) {
      $data.=$chunk;                                  # read the data into var                         
    }
    close (BIN);
    my $lastbyte=chop $data;                          # cut off the checksum
    my $prelastbyte=chop $data;                       
    my $sumnum=unpack 'v',($prelastbyte.$lastbyte);   # convert the binary checksum into a number
    my $sum=unpack("%16C*",$data) % 32767;            # calculate the checksum for the data
    (my $dummy,$stepwidth,$ConfigL::servores,$ConfigL::num_servos)=unpack"vvvv",($data); # read the header into vars  
    $stepwidth*=1000000;                              # convert step duration from ms into ns
    $data=substr $data,8;                             # shorten the file
    my $j=0;                   
    while ( $data ) {                                 # as long as we have content, read it into array
      for (my $i=0;$i<$ConfigL::num_servos;$i++){
        $servocontent[$j][$i]=unpack "v",$data;
        $data=substr $data,2;
      }
      $j++;         
    }
    $contentcount=$j-1;                               # note the number of sets in the file
    $previouscount=$contentcount;                     # remember the number of sets
    printstatus( "Existing $outfilename has $j moves\n");  
    $nummoves=$j;
    if ($sumnum!=$sum) {                              # if checksum is correct, read the data into vars
      printstatus( "loaded file $outfilename has a bad checksum\n");
    }
    $saved=1;
  } else { printstatus("No $outfilename exists - created when saved\n");
    $saved=0;
  }
}
#############################################################
# set the movement range to maximum
#############################################################
sub maxrange{
  $startset->text(0);
  $stopset->text($nummoves);
}
#############################################################
# save the servo file
#############################################################
sub savesv {
  $outfilename=$fileentry->get(); 
  save_file();
  $saved=1;
}
#############################################################
# switch to previous or next mp3 file
#############################################################
sub mpnext {
  mpfile(1);
}
sub mpnext10 {
  mpfile(10);
}
sub mpprev {
  mpfile(-1);
}
sub mpprev10 {
  mpfile(-10);
}
########################################################
# prepare MP3-read
########################################################
sub mpfile {
  my $step=shift;
  if ( $saved==0 ) {
    my $answer=$cui->dialog(-message => "$outfilename has not been saved !\ncontinue without saving?",
                            -buttons=>[
                            { -label => 'No', -value=>0, -shortcut=>'n'},
                            { -label => 'Yes', -value=>1, -shortcut=>'y'},
                            { -label => 'Save', -value=>2, -shortcut=>'s'}
                            ]);
    if ( $answer==0) {
      return;
    }
    if ( $answer==2) {
      savesv();
    }
  }
  $mp3=get_mp($step);
  load_file();
  if ( $maximize->get()==1 ){
    &maxrange;
  }
}
#############################################################
# output path to file section
#############################################################
sub err {
  my $msg=shift;
  $cui->error($msg);
  exit_dialog();
}
#############################################################
# output path to file section
#############################################################
sub showpath {
  $fileentry->text(shift); 
  $fileentry->draw(1);
}
#############################################################
# output text to status section
#############################################################
sub printstatus {
  if ($statuspos >= $statusheight) {
    $statuspos=0; 
  }
  $statuslines[$statuspos]=shift;
  $statuspos++;
  my $statusline='';
  for (my $a=$statuspos;$a<$statusheight;$a++) {
    $statusline.=$statuslines[$a];
  }
  for (my $a=0;$a<$statuspos;$a++){
    $statusline.=$statuslines[$a];
  }
  $status->text($statusline); 
  $status->draw(1);
}
#############################################################
#  make sure never x and y are checked together
#############################################################
sub uncheckx {
  for ($a=0;$a<$ConfigL::num_servos;$a++) {
    if($channely[$a]->get()){
      $channelx[$a]->uncheck();
    }
  } 
}
sub unchecky {
  for ($a=0;$a<$ConfigL::num_servos;$a++) {
    if($channelx[$a]->get()){
      $channely[$a]->uncheck();
    }
  } 
}
#############################################################
#  saves the recorded data
#############################################################
sub save_file{
  # $stepwidth $servores $num_servos                  # vars in use
  # if ($just_loaded) { return }                        # do not save if data has not been changed
  my $bincontent;                                     # this var receives the binary coded servo positions 
  my $j=0;
  foreach my $set (@servocontent) {                   # put the array elements step by step binary coded into the string
    for (my $i=0;$i<$ConfigL::num_servos;$i++) {               
      $bincontent.=pack "v",($set->[$i]);             # save each servo position into a 16 bit word
    }
    if ($j>=$previouscount) { last }                   # only write till the new end
    $j++;
  }
  my $header=pack "vvvv",($previouscount,$stepwidth/1000000,$ConfigL::servores,$ConfigL::num_servos); # binarize the header elements
  my $sum=pack 'v',(unpack("%16C*",$header.$bincontent) % 32767); # calculate the checksum for the complete data stream
  open (BIN,">$outfilename")||err( "Cannot create output file $outfilename $!");
  binmode BIN;
  print BIN $header.$bincontent.$sum;                 # save header,body and checksum into file
  close (BIN);
  printstatus("File $outfilename written\n");
}
#############################################################
#  reads the serial line as long as we get a string, followed by \n
#  returns the content splittet into an array
#############################################################
sub get_one_read {
  my ($count,$data,$i)=(0,0,0);
  while (! (substr $data,-1 eq "\n")) {
    $serial->write('4');              # sent the command to the arduino to send one set
    hsleep ($waittime_serial);
    ($count,$data)=$serial->read(32);
    if ( $count ) {                   # if we have data, put it into the array
      @serial_content=split / /,$data;
      $serial_content[4]=int($ConfigL::servores * ($serial_content[0]-$ConfigL::joystick_x_start) / ($ConfigL::joystick_x_end-$ConfigL::joystick_x_start));
      if ( $serial_content[4] <= 0 ) { $serial_content[4]=1; }
      $serial_content[5]=int($ConfigL::servores * ($serial_content[1]-$ConfigL::joystick_y_start) / ($ConfigL::joystick_y_end-$ConfigL::joystick_y_start));
      if ( $serial_content[5] <= 0 ) { $serial_content[5]=1; }
    } else { $serial->purge_all(); }  # else try again
    if ($i >= 10) { err "serial device does not respond";}
    $i++;
  }
}
#############################################################
#  move one set - send positions to servos
#############################################################
sub put_one_move {
  my $setref=shift;                                   # contains the refereence to one anonymous array containing the moves for one set
  my $i=0;
  foreach my $servopos (@$setref) {                   # set the positions for all of the servos
    my $way=$$ConfigL::servosettings[$i][1]-$$ConfigL::servosettings[$i][0];  # the drive way (resolution) of the servo (205 steps)
    my $resfactor=$way/$ConfigL::servores;                     # calculate the correction factor PCA9685 has 4096 steps
    my $pos;
    if ($$ConfigL::servosettings[$i][2]) {
      $pos=int(($ConfigL::servores-$servopos)*$resfactor+$$ConfigL::servosettings[$i][0]); # take the Servostart from ConfigL::servosettings and invert the direction
    } else {
      $pos=int($servopos*$resfactor+$$ConfigL::servosettings[$i][0]);             # take the Servostart from ConfigL::servosettings
    }
    $dev->setChannelPWM($i,0,$pos);
    &$pwm_en(0);                                      # enable the pwm output
    $i++;
  }
}
#############################################################
#  Does a simple hardware test - this routine never stops !!!
#  only to be run on demand for testing purposes
#############################################################
sub hardwaretest {
  $serial->purge_all();                                         # clear the buffers of the serial line 
  while (1) {
    print ">>>>>";
    $serial->write('4');
    hsleep($waittime_serial);
    my ($count,$data)=$serial->read(32);
    if ($count) {
      print $data;
    }
    print "<<<<<";
  }
}
#############################################################
#  get the next mp3 file name to play
#############################################################
sub get_mp {
  my $move=shift;
  $filepos+=$move;                                 # update the position in the array we are reading from
  if ($filepos <= 0) { $filepos=0 }                   # do not go beyond first file
  if ($filepos > $maxfiles) { $filepos=$maxfiles }    # and do not exceed last file
  return $filelist[$filepos];                         # return the mp3 file to load
}
#############################################################
#  initializes the hardware
#############################################################
sub init_ports {
  $serial = Device::SerialPort->new($ConfigL::serialport); # init the serial port at pin 10 (RXD)
  $serial->baudrate(38400);
  $serial->databits(8);
  $serial->stopbits(1);
  $serial->purge_all();
  $serial->rts_active(0);
  $serial->dtr_active(0);
  $serial->purge_all();                              # flush the buffer
# gpio setup
# ----------
  $pwm_en=sub{$api->write_pin(0,shift)};
  #$pwm_en=sub{Gpio::set('PA1',shift)};
  $inext= sub{return ($api->read_pin(1))};
  #$inext= sub{return !(Gpio::get('PD14'))};
  $iprev= sub{return ($api->read_pin('4'))};
  #$iprev= sub{return !(Gpio::get('PC4'))};
  # init input for next file activate pulldown
  $api->pin_mode(1,0);                               # port 1 (pin 12) as input
  $api->pull_up_down(1,1);                           # activate pulldown
  #Gpio::init('PD14',1,2,0,0);
  # init input for previous/stop file activate pulldown
  $api->pin_mode(4,0);                               # port 4 (pin16) as input
  $api->pull_up_down(4,1);                           # activate pulldown
  #Gpio::init('PC4',1,2,0,0);
  # init pwm enable pin 11 ($pwm_en)as output and set to 1 (inactive)
  $api->pin_mode(0,1);                               # port 0 (pin 11) as output
  $api->write_pin(0,1);                              # init port 0 to 1
  #Gpio::init('PA1',0,0,1,1);  
# i2c setup
# ----------
  $dev = Device::PWMGenerator::PCA9685->new(
    I2CBusDevicePath => $ConfigL::i2cport,           # the i2c device we send commands to
    I2CDeviceAddress => $ConfigL::i2c_address,
    debug            => 0,
    frequency        => $ConfigL::i2c_freq,          #Hz
  );
  $dev->enable();
}
